#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <X11/Xlib.h>
#include "nagratv.h"
#include "nagra.h"
#include "chan.h"

/* ----------------------------------------------------------------------- */

StrTab booltab[] = {
    {  0, "no" },
    {  0, "false" },
    {  0, "off" },
    {  1, "yes" },
    {  1, "true" },
    {  1, "on" },
    { -1, NULL }
};

StrTab audiotab[] = {
  { 0, "no" },
  { 1, "normal" },
  { 2, "nagra" },
};

StrTab videotab[] = {
  { 0, "interlace" },
  { 1, "progressive" },
  { 2, "nagra" },
  { 3, "videocrypt" },
};

void int_to_str(char *buf,int c,StrTab *tab)
{
  int i;
  for (i = 0; tab[i].str != NULL; i++) {
    if (tab[i].nr == c) {
      strcpy(buf,tab[i].str);
      return;
    }
  }
  sprintf(buf,"%d",c);
}

int str_to_int(char *str, struct STRTAB *tab)
{
    int i;
    
    if (str[0] >= '0' && str[0] <= '9')
	return atoi(str);
    for (i = 0; tab[i].str != NULL; i++)
	if (0 == strcasecmp(str,tab[i].str))
	    return(tab[i].nr);
    return -1;
}

/* ----------------------------------------------------------------------- */
/* misc common stuff, not only channel related                             */ 

Channel defaults   = { "defaults", NULL,
                       1, 0, 0, 5, 0,
                       32768, 32768, 32768, 32768,
                       VIDEO_TYPE_INTERLACE,
                       AUDIO_TYPE_NORMAL, 0,
                       75 };

/* ----------------------------------------------------------------------- */

int cf2freq(TVInfo *tvi,int channel, int fine)
{
  double freq = 0;
  int    i;
  int chan_tab=tvi->chan_tab;

  for (i = 0; chan_tabs[chan_tab][i].chan_low != 0; i++) {
    if (chan_tabs[chan_tab][i].chan_low  <= channel &&
        chan_tabs[chan_tab][i].chan_high >= channel) {
	    
	    freq = chan_tabs[chan_tab][i].freq_low +
        chan_tabs[chan_tab][i].factor *
        (channel - chan_tabs[chan_tab][i].chan_low);
	    break;
    }
  }
  
  return (int)(freq*16)+fine;
}

/* load table of 256 ints */
int read_table(unsigned char *table,char *filename)
{
  FILE *f;
  int i,tmp;

  f=fopen(filename,"r");
  if (f == NULL) return -1;
  for(i=0;i<256;i++) {
    fscanf(f,"%d",&tmp);
    table[i]=tmp;
  }
  fclose(f);
  return 0;
}


/* read the configuration from the file 'filename' */
int read_config(GrabInfo *gi,TVInfo *tvi,SoundInfo *si,char *filename)
{
    FILE *fp;
    char line[100], tag[32], val[100];
    int i,nr = 0;
    Channel *current;
    Channel  **channels;
    int channel_count;

    channels=NULL;
    channel_count=0;
    tvi->chan_tab=0;
    tvi->nagra_table=NULL;
    current  = &defaults;

    fp = fopen(filename,"r");
    if (NULL == fp) {
      fprintf(stderr,"can't open config file %s\n",filename);
      return -1;
    }
    while (NULL != fgets(line,99,fp)) {
	nr++;
	if (line[0] == '\n' || line[0] == '#' || line[0] == '%')
	    continue;
	if (1 == sscanf(line,"[%99[^]]]",val)) {
	    if (0 == (channel_count % 16))
		if (!channel_count)
		    channels = malloc(sizeof(struct CHANNEL*)*16);
		else
		    channels = realloc(channels,sizeof(struct CHANNEL*)*(channel_count+16));
	    current = channels[channel_count] = malloc(sizeof(struct CHANNEL));
	    channel_count++;
	    memcpy(current,&defaults,sizeof(struct CHANNEL));
	    current->name = strdup(val);
	    continue;
	}
	if (2 != sscanf(line," %31[^= ] = %99[^\n]",tag,val)) {
	    fprintf(stderr,"parse error line %d\n",nr);
	    exit(1);
	}

	if (0 == strcmp(tag,"key"))
	    current->key = strdup(val);
	
	else if (0 == strcmp(tag,"source"))
	    current->source = str_to_int(val,gi->inputs);
	else if (0 == strcmp(tag,"norm"))
	    current->norm = str_to_int(val,gi->norms);
	else if (0 == strcmp(tag,"channel"))
	    current->channel = atoi(val);
	else if (0 == strcmp(tag,"fine"))
	    current->fine = atoi(val);

	else if (0 == strcmp(tag,"color"))
	    current->color = atoi(val);
	else if (0 == strcmp(tag,"bright"))
	    current->bright = atoi(val);
	else if (0 == strcmp(tag,"hue"))
	    current->hue = atoi(val);
	else if (0 == strcmp(tag,"contrast"))
	    current->contrast = atoi(val);

	else if (0 == strcmp(tag,"volume"))
	    current->volume = atoi(val);

	else if (0 == strcmp(tag,"video")) {
	    if ((current->video_type = str_to_int(val,videotab)) == -1) {
        fprintf(stderr,"Bad video option\n");
        exit(1);
      }
      if (current->video_type == VIDEO_TYPE_NAGRA && tvi->nagra_table == NULL) {
        fprintf(stderr,
                "You must load a nagra table with the option 'nagratable' before\n"
                "using this video type\n");
        exit(1);
      }
	} else if (0 == strcmp(tag,"audio")) {
	    if ((current->audio_type = str_to_int(val,audiotab)) == -1) {
        fprintf(stderr,"Bad audio option\n");
        exit(1);
      }
	} else if (0 == strcmp(tag,"grayscale")) {
	    current->gray_scale = str_to_int(val,booltab);
  } else if (0 == channel_count && 0 == strcmp(tag,"nagratable")) {
	    tvi->nagra_table = malloc(256);
      if (read_table(tvi->nagra_table,val) != 0) {
        fprintf(stderr,"table '%s' not found\n",val);
        exit(1);
      }
  } else if (0 == channel_count && 0 == strcmp(tag,"nagradata")) {
      read_nagra_data(val);
  } else if (0 == channel_count && 0 == strcmp(tag,"mixer")) {
      if (mixer_set_device(si,val) != 0) {
        fprintf(stderr,"bad mixer device '%s'\n",val);
      }
  } else if (0 == channel_count && 0 == strcmp(tag,"fullscreen")) {
	    if (sscanf(val,"%d x %d",&tvi->vidmode_width,&tvi->vidmode_height) != 2) {
        fprintf(stderr,"invalid value for %s: %s\n",tag,val);
        tvi->vidmode_width = tvi->vidmode_height = 0;
	    }
  } else if (0 == channel_count && 0 == strcmp(tag,"freqtab")) {
    if (-1 != (i = str_to_int(val,chan_names))) {
      tvi->chan_tab = i;
    } else {
      fprintf(stderr,"Bad Freqtab\n");
    }
	} else {
	    fprintf(stderr,"parse error line %d: unknown tag %s\n",nr,tag);
	    exit(1);
	}
    }
    fclose(fp);


    tvi->channels=channels;
    tvi->channel_count=channel_count;
    
    tvi->cur_color    = -1;
    tvi->cur_bright   = -1;
    tvi->cur_hue      = -1;
    tvi->cur_contrast = -1;
  
    tvi->cur_input = -1;
    tvi->cur_norm = -1;

    tvi->cur_channel  = -1;
    tvi->cur_fine     = -1;

    tvi->cur_video_type = -1;
    tvi->cur_audio_type = -1;

    tvi->cur_volume = 100;

    tvi->full_screen = 0;
    return 0;
}

