/* The values key_start & key_inc are searched by correlation on a few
 * lines & columns of the frame. Please read the licence in the README
 * file before using this code.  */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#include <errno.h>

/* #define PROFILE */

#define YSIZE 287
#define NB_KEYS 32768

/* Number of pixels per line on which the correlation is achieved.
   must be a multiple of 8 */ 
#define CX_COUNT 32

/* Number of couples of lines on which the correlation is
   achieved. You must modify it in the asm source too !!!!!! */
#define NB_COUPLES 40

/* The correlation use only the lines after LINE_MIN */
#define LINE_MIN 96

/* The correlation use only the lines before LINE_MAX */
#define LINE_MAX (286-96)

unsigned char table_keys[256];
unsigned short key_lines[NB_KEYS][NB_COUPLES];
unsigned int couple_index[6000];
int couple_index_nb;
unsigned int correl_lines[6000];
int valid_data=0;
char *nagra_data_file = NULL;

/* init variables */
int nb_couples;
int *couple_next;
int *couple_first;
int *couple_score;
int *couple_nb;

/* compute the permutation generated by the 15 bit key 'key' */
static void gen_perm(unsigned short *perm,int key)
{
  int key_start,key_inc,key_index,i,b;
  int buffer[32];
  key_start=key >> 7;
  key_inc=(key & 0x7f)*2+1;
  for(i=0;i<32;i++) buffer[i]=i;
  key_index=key_start;
  for(i=0;i<255;i++) 
   { 
    b=table_keys[key_index];
    perm[i]=buffer[b];
    buffer[b]=i+32;
    key_index=(key_index+key_inc) & 0xff;
   }
  for(i=0;i<32;i++) 
   perm[255+i]=buffer[i];
}

void add_couple(int c)
{
  int p,key,q,y1,y2,nb;
  y1=c / YSIZE;
  y2=c % YSIZE;
  couple_index[couple_index_nb]=(y1 << 9) | y2;
  p=couple_first[c];
  while (p!=-1) {
    key=p / nb_couples;
    nb=couple_nb[key]++;
    q=couple_next[p];
    if (nb < NB_COUPLES) {
      key_lines[key][nb]=couple_index_nb*4;
    }
    p=q;
  }
  couple_index_nb++;
}
int score_couple(int c)
{
  int p,key,score;
  score=0;
  p=couple_first[c];
  while (p!=-1) {
    key=p / nb_couples;
    if (couple_nb[key] < NB_COUPLES) score++;
    p=couple_next[p];
  }
  return score;
}

static void init_decode_tables(void)
{
  FILE *f;
  int key,i,j,last_y,y,p,q,score;
  unsigned short perm[YSIZE];
  int nb_tot,score_max;
  int line_min=LINE_MIN,line_max=LINE_MAX;
  printf("Initialising nagra decoding tables\n");
  fflush(stdout);
  nb_couples=line_max-line_min;

  couple_next=malloc(sizeof(int)*nb_couples*NB_KEYS);
  couple_first=malloc(YSIZE*YSIZE*sizeof(int));
  couple_score=malloc(YSIZE*YSIZE*sizeof(int));
  couple_nb=malloc(NB_KEYS*sizeof(int));

  for(i=0;i<YSIZE*YSIZE;i++) { couple_first[i]=-1; couple_score[i]=0; }
  for(key=0;key<NB_KEYS;key++) {
    gen_perm(perm,key);
    last_y=perm[line_min];
    for(i=line_min+1;i<=line_max;i++) {
      y=perm[i];
      if (y < last_y) j=y*YSIZE+last_y; else j=last_y*YSIZE+y;
      last_y=y;
      q=(key*nb_couples)+i-(line_min+1);
      p=couple_first[j];
      couple_next[q]=p;
      couple_first[j]=q;
      couple_score[j]++;
    }
    if ((key & 1023) == 0) {
      printf(".");
      fflush(stdout);
    }
  }
  for(i=0;i<NB_KEYS;i++) couple_nb[i]=0;
  nb_tot=0;
  couple_index_nb=0;
  do {
    score_max=0;
    j=0;
    for(i=0;i<YSIZE*YSIZE;i++) {
      if (couple_score[i] > score_max) {
        score_max=couple_score[i];
        j=i;
      }
    }
    score=score_couple(j);
    if (score > 0) {
      nb_tot+=score;
      add_couple(j);
    }
    couple_score[j]=0;
    if ((nb_tot & 63) == 0) { printf(".");fflush(stdout); }
 } while (nb_tot < (NB_KEYS*NB_COUPLES)) ;
  free(couple_nb);
  free(couple_score);
  free(couple_first);
  free(couple_next);
  printf("\n[correlation on %d couples]\n",couple_index_nb);
  if (nagra_data_file != NULL) {
      if ((f=fopen(nagra_data_file,"w")) == NULL) {
	  fprintf(stderr,"Write error nagradata '%s'\n",nagra_data_file);
	  return;
      }
      fprintf(f,"NagraTV 2.3\nNB_COUPLES=%d\ncouple_index_nb=%d\n",
	      NB_COUPLES,couple_index_nb);
      fwrite(couple_index,sizeof(int),couple_index_nb,f);
      fwrite(key_lines,sizeof(short),NB_COUPLES*NB_KEYS,f);
      fclose(f);
  }
}

void init_decode_nagra(unsigned char *table)
{
    memcpy(table_keys,table,256);
    if (valid_data == 0)
	init_decode_tables();
}

void read_nagra_data(char *filename)
{
    FILE *f;
    int i;
    
    if ((f=fopen(filename,"r")) == NULL) {
	if (errno == ENOENT) {
	    fprintf(stderr,"Nagradata '%s' not found\n",filename);
	    nagra_data_file = malloc(strlen(filename)+1);
	    strcpy(nagra_data_file,filename);
	    return;
	} else {
	    perror(filename);
	}
    }
    if (fscanf(f,"NagraTV 2.3\nNB_COUPLES=%d\ncouple_index_nb=%d\ncouple_index=\n",
	       &i,&couple_index_nb) != 2) {
	fprintf(stderr,"Nagradata '%s' is corrupted!\nPlease delete the file and create a new one.\n",filename);
	exit(1);
    }
    if (i != NB_COUPLES) {
	fprintf(stderr,"Nagradata '%s' has different NB_COUPLES!\nPlease delete the file and create a new one.\n",filename);
	exit(1);
    }
    
    printf("Read nagradata '%s':\nNB_COUPLES=%d\ncouples_index_nb=%d\n",filename,i,
	   couple_index_nb);
  
    if ((fread(couple_index,sizeof(int),couple_index_nb,f) != couple_index_nb) ||
	(fread(key_lines,sizeof(short),NB_COUPLES*NB_KEYS,f) != NB_COUPLES*NB_KEYS)) {
	fprintf(stderr,"Nagradata '%s' is corrupted!\nPlease delete the file and create a new one.\n",filename);
	exit(1);
    }
    fclose(f);
    valid_data = 1;
    nagra_data_file = malloc(strlen(filename)+1);
    strcpy(nagra_data_file,filename);
}

unsigned short *try_perm1(void);
void correlation(unsigned short *);

/* the following variables are used to look at the 25 frame key
   repetition. It is for information purpose only since it is not used
   in the current code. */

#define KEY_BUF_SIZE 25
int key_locked;
int key_buf_ptr=0;
int key_buf[KEY_BUF_SIZE];

int decode_nagra(unsigned short *perm,
	void **lines,int xsize,int format)
{
    int i,j;
    unsigned short pts_buffer[YSIZE][CX_COUNT],*p;
    int key;
    unsigned short *key_ptr;
    int cx_inc,cx_start;

#ifdef PROFILE
    int ti=clock();
#endif

    cx_inc=xsize/CX_COUNT;
    cx_start=(cx_inc/2)&0x7e;

    p=&pts_buffer[0][0];

    /* preload of the pixels used for the correlation */
    switch(format) {
	case 1:
	    /* Y format */
	    for(i=0;i<YSIZE;i++) {
		unsigned char *q;

		q=((unsigned char **)lines)[i] + cx_start;
		for(j=0;j<CX_COUNT;j+=4) {
		    p[0]=q[0];
		    q+=cx_inc;
		    p[1]=q[0];
		    q+=cx_inc;
		    p[2]=q[0];
		    q+=cx_inc;
		    p[3]=q[0];
		    q+=cx_inc;
		    p+=4;
		}
	    }
	    break;
	case 2:
	    /* YUY2 format */
	    cx_inc += cx_inc;
	    for(i=0;i<YSIZE;i++) {
		unsigned char *q;

		q=((unsigned char **)lines)[i] + cx_start;
		for(j=0;j<CX_COUNT;j+=4) {
		    p[0]=q[0];
		    q+=cx_inc;
		    p[1]=q[0];
		    q+=cx_inc;
		    p[2]=q[0];
		    q+=cx_inc;
		    p[3]=q[0];
		    q+=cx_inc;
		    p+=4;
		}
	    }
	    break;
    }

    /* Precompute the correlation between the lines, asm */
    correlation(&pts_buffer[0][0]);

    key_ptr=try_perm1();
    key=(key_ptr - (unsigned short *)key_lines) / NB_COUPLES;

    gen_perm(perm,key);

    if (key_buf[key_buf_ptr] == key) {
	key_locked='*';
    } else {
	key_locked=' ';
    }
    key_buf[key_buf_ptr++]=key;
    if (key_buf_ptr == KEY_BUF_SIZE) key_buf_ptr=0;

#ifdef PROFILE
    ti=clock() - ti;
    fprintf(stderr,"key: %04X time=%0.3fs\n",key,
	    (double) ti / CLOCKS_PER_SEC);
#endif

    return 0;
}

