#include <stdlib.h>
#include <stdio.h>

#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <signal.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <time.h>
#include <string.h>
#include <sched.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/extensions/xf86dga.h>
#include <X11/extensions/xf86vmode.h>

#include <videodev.h>
#include <bttv.h>

#include "nagra.h"
#include "vcrypt.h"
#include "nagratv.h"

/* #define PROFILE */

/* audio processing stuff */

void catch_children()
{
	while (wait3 (NULL, WNOHANG, NULL) > 0);
}

int audio_pid=-1;
void audio_decode_end(void)
{
	if (audio_pid != -1) {
		kill(audio_pid,SIGQUIT);
	}
	audio_pid=-1;
}

void audio_decode_start(SoundInfo *si,Display *dpy)
{
	/* do the wait automatically */
	signal(SIGCHLD,catch_children);

	/* kill the process at the end */
	atexit(audio_decode_end);

	audio_pid=fork();
	if (audio_pid == -1) {
		perror("fork");
		exit(1);
	}
	if (audio_pid > 0) return;

	mixer_set_decode(si);
	do_audio_decode();
}


/* video decode stuff */

unsigned short y_to_rgb16[256];
unsigned char *lines_addr[2][287];
int picture=0;
int phase=0;
int bpp=0;
int bars=0;
int interlace=0;

extern void putline0(void* src, void* dest, int size);
extern void putline1(void* src, void* dest, int size);
extern void putline2(void* src, void* dest, int size);
extern void putline3(void* src, void* dest, int size);

extern void putlinenocolor0(void* src_y, void* src_uv, void* dest, int size);
extern void putlinenocolor1(void* src_y, void* src_uv, void* dest, int size);
extern void putlinenocolor2(void* src_y, void* src_uv, void* dest, int size);
extern void putlinenocolor3(void* src_y, void* src_uv, void* dest, int size);

#define color1 \
diff = (perm[y]-y)&3; \
color();

#define color2 \
diff = (perm[y]-y)&3; \
if (diff==1) diff=3; \
else if (diff==3) diff=1; \
color();

void decode_frame_nagra(unsigned char *dest,int dest_xsize,
		unsigned char **lines,int xsize,int ysize,
		int gray_scale,int dup_lines)
{
    int y,x,format;
    unsigned short perm[288]; /* statt 287 */
    short *p2;
    unsigned char *p1;
    unsigned char *pc;
    int diff;
    int decode;

    static inline void no_color(void) { 
	switch (diff) { 
	case 0: 
	    putlinenocolor0((void*)p1,(void*)pc,(void*)p2,xsize); 
	    break; 
	case 1: 
	    putlinenocolor1((void*)p1,(void*)pc,(void*)p2,xsize); 
	    break; 
	case 2: 
	    putlinenocolor2((void*)p1,(void*)pc,(void*)p2,xsize); 
	    break; 
	case 3: 
	    putlinenocolor3((void*)p1,(void*)pc,(void*)p2,xsize); 
	    break; 
	} 
    }

    static inline void color(void) { 
	pc=p1; 
	switch (diff) { 
	case 0: 
	    putline0((void*)p1,(void*)p2,xsize); 
	    break; 
	case 1: 
	    putline1((void*)p1,(void*)p2,xsize); 
	    break; 
	case 2: 
	    putline2((void*)p1,(void*)p2,xsize); 
	    break; 
	case 3: 
	    putline3((void*)p1,(void*)p2,xsize); 
	    break; 
	} 
    }



    /* find the permutation  */
    format=2;
    if (gray_scale) format=1;

    decode = decode_nagra(perm,(void **)lines,xsize,format);

    /* do the permutation */
    if (gray_scale) {
	p2=(short *)dest;
	for(y=bars;y<287-bars;y++) {
	    p1=lines[y];
	    if (decode==0) p1=lines[perm[y]];
	    for(x=0;x<xsize;x+=4) {
		p2[0]=y_to_rgb16[p1[0]];
		p2[1]=y_to_rgb16[p1[1]];
		p2[2]=y_to_rgb16[p1[2]];
		p2[3]=y_to_rgb16[p1[3]];
		p2+=4;
		p1+=4;
	    }
	    p2+=dest_xsize-xsize;

	    if (dup_lines) {
		p1-=xsize;
		for(x=0;x<xsize;x+=4) {
		    p2[0]=y_to_rgb16[p1[0]];
		    p2[1]=y_to_rgb16[p1[1]];
		    p2[2]=y_to_rgb16[p1[2]];
		    p2[3]=y_to_rgb16[p1[3]];
		    p2+=4;
		    p1+=4;
		}
		p2+=dest_xsize-xsize;
	    }
	}
    } else { 
	if (phase) picture+=2;
	p2 = (short *) dest+bars*dest_xsize;;
	pc = lines[perm[0]];
	diff = 0;
	for (y = bars; y < 287-bars; y++)
	    {
		p1 = lines[perm[y]];
		if (perm[y]<33) { no_color(); } 
		else {
		    if (((32+y-perm[y])>35)||((32+y-perm[y])<5)) { no_color(); }
		    else {
			if ((picture&2)==0) {
			    if ((perm[y] & 1)==0) { color1; }
			    else   { color2; } 
			}
			else {
			    if ((perm[y] & 1)==1) { color1; }
			    else  { color2; } 
			}
		    }
		}
		p2 += dest_xsize;
		if (dup_lines) {
		    if (!interlace) {
			if (perm[y]<33) { no_color(); } 
			else {
			    if (((32+y-perm[y])>35)||((32+y-perm[y])<5)) { no_color(); }
			    else {
				if ((picture&2)==0) {
				    if ((perm[y] & 1)==0) { color1; }
				    else   { color2; } 
				}
				else {
				    if ((perm[y] & 1)==1) { color1; }
				    else  { color2; } 
				}
			    }
			}
		    }
		    p2+=dest_xsize;
		}
	    } 
	if (phase) picture-=2;
    }
}

void decode_frame_vc (unsigned char *dest, int dest_xsize,
		      unsigned char *src, int xsize, int ysize)
{
    int y, x;
    short *p2;
    unsigned char *p1;
    
    if (xsize == 384) {
	decodeFrameVC(src,0);
	p2 = (short *) dest+bars*dest_xsize;
	for (y = bars; y < 288-bars; y++) {
	    p1 = &src[y*xsize*4];
	    for (x = 0; x < (xsize/2); x++) {
		((unsigned int *) p2)[0] = y_to_rgb16[p1[0]] | ((y_to_rgb16[p1[2]]) << 16);
		p2 += 2;
		p1 += 4;
	    }
	    p2 += dest_xsize - xsize;
	}
    } else {
	decodeFrameVCFullRes(src,0);

        p2 = (short *) dest+bars*dest_xsize;
	if (interlace) {
	    for (y = bars; y < 288-bars; y++) {
		p1 = &src[y*xsize*4];
		for (x = 0; x < (xsize/2); x++) {
		    ((unsigned int *) p2)[0] = y_to_rgb16[p1[0]] | ((y_to_rgb16[p1[2]]) << 16);
		    p2 += 2;
		    p1 += 4;
		}
		p2 += 2*dest_xsize - xsize;
	    }
	} else {
	    for (y = bars; y < 288-bars; y++) {
		p1 = &src[y*xsize*4];
		for (x = 0; x < (xsize/2); x++) {
		    ((unsigned int *) p2)[0] = 
		    ((unsigned int *) (p2+dest_xsize))[0] = 
                         y_to_rgb16[p1[0]] | ((y_to_rgb16[p1[2]]) << 16);

		    p2 += 2;
		    p1 += 4;
		}
		p2 += 2*dest_xsize - xsize;
	    }
	}
	
    }
}


void copy_interlace(unsigned char *dest,int dest_xsize,
		unsigned char *src,int xsize,int ysize)
{
	int y;
	register short *p2;
	register unsigned char *p1;

	p2=(short *)dest;
	p1=src;

	for(y=0;y<ysize;y+=2) {
		memcpy(p2,p1,2*xsize);
		p2+=dest_xsize;

		memcpy(p2,p1,2*xsize);
		p2+=dest_xsize;

		p1+=4*xsize;
	}
}

void copy_interlace_small(unsigned char *dest,int dest_xsize,
		unsigned char *src,
		int xsize,int ysize)
{
	int y;
	register short *p2;
	register unsigned char *p1;

	p2=(short *)dest;
	p1=src;

	for(y=0;y<ysize;y+=2) {
		memcpy(p2,p1,2*xsize);
		p2+=dest_xsize;
		p1+=4*xsize;
	}
}

void copy_progressive(unsigned char *dest,int dest_xsize,
		unsigned char *src,int xsize,int ysize)
{
	int y;
	register short *p2;
	register unsigned char *p1;

	p2=(short *)dest;
	p1=src;
	for(y=0;y<ysize;y++) {
		memcpy(p2,p1,xsize*2);
		p1+=2*xsize;
		p2+=dest_xsize;
	}
}



int debug;

void set_video_type(DispInfo *di,GrabInfo *gi,TVInfo *tvi,int video_type)
{
	struct video_mmap *gb;
	int i,frame,xsize1;
	unsigned char *src;

	ClearScreen(di);
	tvi->cur_video_type=video_type;

	/* init grabbing structure */
	gb=tvi->gb;
	if (bpp==15)
	gb->format = BT848_COLOR_FMT_RGB15;
	else
	gb->format = BT848_COLOR_FMT_RGB16;	

	if (tvi->cur_video_type == VIDEO_TYPE_NAGRA) 
		gb->format = BT848_COLOR_FMT_YUY2;

	if (tvi->cur_video_type == VIDEO_TYPE_VCRYPT) 
		gb->format = BT848_COLOR_FMT_YUY2;

	if (tvi->cur_gray_scale && tvi->cur_video_type == VIDEO_TYPE_NAGRA)
		gb->format = BT848_COLOR_FMT_Y8;

	if (!tvi->full_screen) {
		tvi->xsize=TV_XSIZE/2;
		tvi->ysize=TV_YSIZE/2;
	} else {
		tvi->xsize=TV_XSIZE;
		tvi->ysize=TV_YSIZE;
	}

	tvi->posx=di->vp_width/2-tvi->xsize/2;
	tvi->posy=di->height/2-tvi->ysize/2;
	if (tvi->posx < 0) tvi->posx=0;
	if (tvi->posy < 0) tvi->posy=0;

	gb->width  = tvi->xsize;
	gb->height = TV_YSIZE; /* we take the whole frame in both cases */

	tvi->dst=di->base+(di->width*tvi->posy+tvi->posx)*2;

	/* build the nagra line address table */
	xsize1=tvi->xsize;
	
	if (gb->format == BT848_COLOR_FMT_RGB16) xsize1=2*xsize1;
	if (gb->format == BT848_COLOR_FMT_YUY2)  xsize1=2*xsize1;

	for(frame=0;frame<2;frame++) {
		if (frame == 0) src=gi->buffer0;
		else src=gi->buffer1;
		for(i=0;i<32;i++) lines_addr[frame][i]=&src[(2*(288-33+i))*(xsize1)];
		for(i=0;i<255;i++) lines_addr[frame][i+32]=&src[(2*i+1)*(xsize1)];
	}
}

void set_audio_type(DispInfo *di,TVInfo *tvi,SoundInfo *si,int audio_type)
{
	if (audio_type != tvi->cur_audio_type) {
		if (audio_type == AUDIO_TYPE_NAGRA) {
			audio_decode_start(si,di->dpy);
		} else if (tvi->cur_audio_type == AUDIO_TYPE_NAGRA) {
			audio_decode_end();
		}

		if (audio_type == AUDIO_TYPE_NORMAL) {
			mixer_set_normal(si);
		}

		tvi->cur_audio_type=audio_type;
	}
}

/* change the current channel */
void set_channel(GrabInfo *gi,TVInfo *tvi,DispInfo *di,SoundInfo *si,int c)
{
	Channel *channel;

	channel=tvi->channels[c];
	tvi->cur_channel_num=c;

	/* picture params */
	tvi->cur_color    = channel->color;
	tvi->cur_bright   = channel->bright;
	tvi->cur_hue      = channel->hue;
	tvi->cur_contrast = channel->contrast;
	grab_picture(gi,tvi->cur_color,tvi->cur_bright,tvi->cur_hue,tvi->cur_contrast);

	/* norm & input */
	if (tvi->cur_input != channel->source || tvi->cur_norm != channel->norm) {
		grab_input(gi,channel->source,channel->norm);
		tvi->cur_norm     = channel->norm;
		tvi->cur_input    = channel->source;
	}

	/* tuning */
	tvi->cur_channel  = channel->channel;
	tvi->cur_fine     = channel->fine;
	grab_tune(gi,cf2freq(tvi,channel->channel,channel->fine));

	/* video & audio processing */

	set_audio_type(di,tvi,si,channel->audio_type);

	tvi->cur_gray_scale = channel->gray_scale;
	set_video_type(di,gi,tvi,channel->video_type);

	/* volume */
	if (channel->volume != tvi->cur_volume) {
		tvi->cur_volume=channel->volume;
		mixer_set_volume(si,tvi->cur_volume);
	}
}

/* search to which channel the key 'keysym' corresponds */
int test_channel_key(TVInfo *tvi,int keysym)
{
	int i;
	char *key_str;

	key_str=XKeysymToString(keysym);
	for(i=0;i<tvi->channel_count;i++) {
		if (strcasecmp(key_str,tvi->channels[i]->key) == 0) {
			return i;
		}
	}
	return -1;
}

#define SPEED_COUNT_MAX 10

static double speed_last_time,speed_fps;
static int speed_count=0;

extern int key_locked;

void DisplayStatus(DispInfo *d,TVInfo *tvi)
{
	Channel *channel;
	
        char buf[500],buf1[20],buf2[20];
	int freq;
	struct timeval tv;
	double time1;

	speed_count++;
	if (speed_count == SPEED_COUNT_MAX) {
		gettimeofday(&tv,NULL);
		time1 = tv.tv_sec + tv.tv_usec*1E-6;
		speed_fps = SPEED_COUNT_MAX / (time1 - speed_last_time);
		speed_last_time=time1;
		speed_count=0;
}

	channel=tvi->channels[tvi->cur_channel_num];
        freq=cf2freq(tvi,tvi->cur_channel,tvi->cur_fine); 
        int_to_str(buf1,tvi->cur_video_type,videotab);
	int_to_str(buf2,tvi->cur_audio_type,audiotab);
	/* channel info */
	sprintf(buf,"%-8s nr:%3d  cor:%3d  %6.2fMHz  mix =%3d",channel->name,tvi->cur_channel,tvi->cur_fine,(double) freq / 16.0,tvi->cur_volume);


	DisplayText(d,tvi->posx+2*FONT_XSIZE,tvi->posy+FONT_YSIZE,buf,strlen(buf));

	sprintf(buf,
			"fps=%4.1f  video=%-11s  audio=%-10s%c",
			speed_fps,buf1,buf2,key_locked);

	DisplayText(d,tvi->posx+2*FONT_XSIZE,tvi->posy+tvi->ysize-2*FONT_YSIZE,buf,strlen(buf));
}


/* change the current X11 video mode */

int vm_count,vm_smode;
XF86VidModeModeInfo **vm_modelines;

void video_mode_init(TVInfo *tvi,Display *dpy)
{
	int i;

	vm_smode=-1;

	fprintf(stderr,"x11 modes:\n");
	XF86VidModeGetAllModeLines(dpy,XDefaultScreen(dpy),
			&vm_count,&vm_modelines);

	for (i = 0; i < vm_count; i++) {

		if (vm_modelines[i]->hdisplay == tvi->vidmode_width &&
				vm_modelines[i]->vdisplay == tvi->vidmode_height) {
			vm_smode=i;
		}

		if (i == 0) {
			fprintf(stderr,"\t%d: %dx%d (current mode)\n",i,vm_modelines[i]->hdisplay,vm_modelines[i]->vdisplay);
		} else {
			fprintf(stderr,"\t%d: %dx%d\n",i,vm_modelines[i]->hdisplay,vm_modelines[i]->vdisplay);
		}
	}

	if (vm_smode != -1) {
		fprintf(stderr,"\tswitching to modeline: %d\n",vm_smode);
		XF86VidModeSwitchToMode(dpy,XDefaultScreen(dpy),vm_modelines[vm_smode]);
		XFlush(dpy);
	}
}

void video_mode_close(TVInfo *tvi,Display *dpy)
{
	if (vm_smode != -1) {
		XF86VidModeSwitchToMode(dpy,XDefaultScreen(dpy),vm_modelines[0]);
		XFlush(dpy);
	}
}

#define STATUS_DELAY 30

int main(int argc,char **argv) 
{
	char bttvname[1024]="/dev/bttv";
	char config_file[1024];
	unsigned char *addr;
	int sbpp,image_size;
	int flags;
	int i;
	int vp_width,vp_height,bank,ram;
	Display *dpy;
	int EventBase, ErrorBase;
	XEvent xev;
	KeySym keysym;
	XComposeStatus status;
	XWindowAttributes wts;
	char buf[80];
	int disp_status_delay,disp_status_toggle,buffer_nb,c;
	struct video_mmap gb;
	GrabInfo gi;
	TVInfo tvi;
	DispInfo di;
	SoundInfo si;

#ifdef PROFILE
	int ti;
#endif

	/* command line options */
	debug=0;
	sprintf(config_file,"%s/%s",getenv("HOME"),".nagratvrc");

	while (1) {
		c=getopt(argc,argv,"shvd:r:");
		if (c == -1) break;
		switch(c) {
			case 'h':
				printf("NagraTV version 2.1 by Gaston
					with PAL-color by Stefan\n"
						"usage: nagratv [-s] [-h] [-v] [-d device] [-r config]\n"
						"\n"
						"-s : set scheduler priority high\n"
						"-r config : read the config file from 'config'\n"
						"-d device : use the video4linux device 'device'\n"
						"-v : verbose mode (debug)\n"
						"-h : help\n"
						"During decoding, press 'q' or 'ESC' to quit.\n"
						"Read the README file for more information.\n"
				      );
				exit(1);
			case 's':
				sync();
				mlockall (MCL_CURRENT|MCL_FUTURE);
				{ struct sched_param sp;
				  sp.sched_priority=99;
				  sched_setscheduler(0,SCHED_RR,&sp); }
				break;		
			case 'd':
				strcpy(bttvname,optarg);
				break;
			case 'v':
				debug=1;
				break;
			case 'r':
				strcpy(config_file,optarg);
				break;
		}
	}

	sbpp=2;
	image_size=TV_XSIZE*TV_YSIZE*sbpp;
	tvi.gb=&gb;

	/* open the grabber */
	if (grab_open(&gi,bttvname,TV_XSIZE,TV_YSIZE,sbpp) != 0) {
		fprintf(stderr,"Error while opening the video device '%s'. Exiting\n",bttvname);
		exit(1);
	}

	/* open the mixer */
	mixer_open(&si,"/dev/mixer");

	/* read the configuration */
	if (read_config(&gi,&tvi,&si,config_file) != 0) {
		fprintf(stderr,"Error while reading the config file '%s'. Exiting\n",config_file);
		exit(1);
	}

	/* open the X11 display */
	dpy=XOpenDisplay("");
	if (dpy==NULL)   {
		fprintf(stderr,"Can't open display\n");
		exit(-2);
	}

	/* initialize the nagra decoder */
	if (tvi.nagra_table != NULL) {
		init_decode_nagra(tvi.nagra_table);
	}

	/* change the current video mode if needed */
	video_mode_init(&tvi,dpy);

	XGetWindowAttributes(dpy, DefaultRootWindow(dpy), &wts);
	bpp=wts.depth;
	if (bpp !=16 && bpp !=15)
	{ fprintf(stderr,"This program only runs with a color depth of 15 or 16 bits.\nCurrent color depth is %d bpp.\n",bpp);
	exit(1);
	}

	if (XF86DGAQueryExtension(dpy,&EventBase,&ErrorBase)) {
		XF86DGAQueryDirectVideo(dpy,XDefaultScreen(dpy),&flags);
		if ( (flags & XF86DGADirectPresent) == 0) {
			fprintf(stderr,"dga: no direct present, exiting\n");
			exit(1);
		}
	}

	/* grab the keyboard */
	XGrabKeyboard(dpy, DefaultRootWindow(dpy), True, GrabModeAsync,
			GrabModeAsync,  CurrentTime);

	/* and all the mouse moves */
	XGrabPointer(dpy, DefaultRootWindow(dpy), True,
			PointerMotionMask | ButtonPressMask | ButtonReleaseMask,
			GrabModeAsync, GrabModeAsync, None,  None, CurrentTime);


	XF86DGAGetVideo(dpy,XDefaultScreen(dpy),(char **)&di.base,
			&di.width,&bank,&ram);
	if (debug) {
		fprintf(stderr,"dga: base=%p width=%d bank=%d ram=%d\n",
				di.base,di.width,bank,di.width);
	}

	XF86DGAGetViewPortSize(dpy,XDefaultScreen(dpy),
			&vp_width, &vp_height);
	di.vp_width=vp_width;
	di.height=vp_height;
	di.dpy=dpy;

	if (debug) {
		fprintf(stderr,"dga: viewport size: %dx%d\n",
				vp_width,vp_height);
	}

	/* move the viewport to the origin */
	XF86DGASetViewPort(dpy,XDefaultScreen(dpy),0,0);

	XF86DGADirectVideo(dpy,XDefaultScreen(dpy),
			XF86DGADirectGraphics | 
			XF86DGADirectMouse | 
			XF86DGADirectKeyb );

	for(i=0;i<256;i++) 
	{ if (bpp==15) y_to_rgb16[i]=(i>>3) | ((i>>3) <<5) | ((i>>3) <<10);
	  else  
	  y_to_rgb16[i]=(i>>3) | ((i>>2) <<5) | ((i>>3) <<11);
	}

	/* set the first channel */
	set_channel(&gi,&tvi,&di,&si,0);

	ClearScreen(&di);
	buffer_nb=0;
	disp_status_delay=STATUS_DELAY;
	disp_status_toggle=0;

	while (1) {
		/* X events : we wait for a key */
		if (XPending(dpy) > 0) {
			XNextEvent(dpy,&xev);
			switch(xev.type) {
				case KeyPress:
					disp_status_delay=STATUS_DELAY;
					XLookupString((XKeyEvent *)&xev,buf,80,&keysym,&status);
					/* user defined keys to change the channel */
					c=test_channel_key(&tvi,keysym);
					if (c != -1) {
set_channel(&gi,&tvi,&di,&si,c);
					} else {
						switch(keysym) {
case XK_Up:
/* previous channel */
if (tvi.cur_channel <32) /* Hier erhoehen, falls erforderlich, siehe chan.h */
	{
	tvi.cur_channel++;
	grab_tune(&gi,cf2freq(&tvi,tvi.cur_channel,tvi.cur_fine));
	}
break;

case XK_Down:
/* next channel */
if (tvi.cur_channel >1) 
	{
	tvi.cur_channel--;
	grab_tune(&gi,cf2freq(&tvi,tvi.cur_channel,tvi.cur_fine));
	}
break;

case XK_Left:
/* fine tuning */
tvi.cur_fine--;
grab_tune(&gi,cf2freq(&tvi,tvi.cur_channel,tvi.cur_fine));
break;

case XK_Right:
/* fine tuning */
tvi.cur_fine++;
grab_tune(&gi,cf2freq(&tvi,tvi.cur_channel,tvi.cur_fine));
break;

case XK_s:
/* display status toggle */
disp_status_toggle^=1;
disp_status_delay=0;
break;

case XK_BackSpace:
/* initial V-phase */
phase=1-phase;
disp_status_delay=0;
break;

/* Blanking */

case XK_1:
bars=0;
break;

case XK_2:
bars=33;
disp_status_delay=0;
ClearScreen(&di); 
break;

case XK_3:
bars=38;
disp_status_delay=0;
ClearScreen(&di);
break;

case XK_4:
bars=43;
disp_status_delay=0;
ClearScreen(&di);
break;

case XK_5:
bars=48;
disp_status_delay=0;
ClearScreen(&di);
break;

case XK_6:
bars=53;
disp_status_delay=0;
ClearScreen(&di);
break;

case XK_7:
bars=58;
disp_status_delay=0;
ClearScreen(&di);
break;

case XK_8:
bars=63;
disp_status_delay=0;
ClearScreen(&di);
break;

case XK_9:
bars=67;
disp_status_delay=0;
ClearScreen(&di);
break;



/* video type */
case XK_p:
set_video_type(&di,&gi,&tvi,VIDEO_TYPE_PROGRESSIVE);
break;
case XK_i:
if ((tvi.cur_video_type == VIDEO_TYPE_PROGRESSIVE) ||
    (tvi.cur_video_type == VIDEO_TYPE_INTERLACE))
    set_video_type(&di,&gi,&tvi,VIDEO_TYPE_INTERLACE);
else {
    interlace^=1;
    ClearScreen(&di);
}
break;
case XK_n:
set_video_type(&di,&gi,&tvi,VIDEO_TYPE_NAGRA);
break;
case XK_v:
set_video_type(&di,&gi,&tvi,VIDEO_TYPE_VCRYPT);
break;

/* volume */
case XK_Next:
tvi.cur_volume-=1;
if (tvi.cur_volume < 0) tvi.cur_volume=100;
mixer_set_volume(&si,tvi.cur_volume);
break;

case XK_Prior:
tvi.cur_volume+=1;
if (tvi.cur_volume > 100) tvi.cur_volume=100;
mixer_set_volume(&si,tvi.cur_volume);
break;


							case XK_f:
								/* full screen toggle */
								tvi.full_screen^=1;
								set_video_type(&di,&gi,&tvi,tvi.cur_video_type);
								break;

								/* gray scale */
							case XK_g:
								tvi.cur_gray_scale^=1;
								set_video_type(&di,&gi,&tvi,tvi.cur_video_type);
								break;

								/* audio type */
							case XK_a:
								set_audio_type(&di,&tvi,&si,AUDIO_TYPE_NORMAL);
								break;
							case XK_z:
								set_audio_type(&di,&tvi,&si,AUDIO_TYPE_NAGRA);
								break;

							case XK_q:
							case XK_Escape:
								/* exit */
								goto the_end;
						}
					}
					break;
			}
		}

		/* we use a double buffer */

		tvi.gb->frame=buffer_nb;

		picture=grab_frame(&gi,tvi.gb);

		if (buffer_nb != 0) addr=gi.buffer0; else addr=gi.buffer1;

		/* compute screen address */

#ifdef PROFILE
		ti=clock();
#endif

		if (tvi.full_screen) {
			/* big image */
			switch(tvi.cur_video_type) {
				case VIDEO_TYPE_NAGRA:
					decode_frame_nagra(tvi.dst,di.width,&lines_addr[1-buffer_nb][0],
							TV_XSIZE,TV_YSIZE,tvi.cur_gray_scale,1);
					break;
				case VIDEO_TYPE_VCRYPT:
					decode_frame_vc(tvi.dst,di.width,addr,TV_XSIZE,TV_YSIZE);
					break;
				case VIDEO_TYPE_INTERLACE:
					copy_interlace(tvi.dst,di.width,addr,TV_XSIZE,TV_YSIZE);
					break;
				case VIDEO_TYPE_PROGRESSIVE:
					copy_progressive(tvi.dst,di.width,addr,TV_XSIZE,TV_YSIZE);
					break;
			}
		} else {
			/* small image : field only display */
			switch(tvi.cur_video_type) {
				case VIDEO_TYPE_NAGRA:
					decode_frame_nagra(tvi.dst,di.width,&lines_addr[1-buffer_nb][0],
							TV_XSIZE/2,TV_YSIZE,tvi.cur_gray_scale,0);
					break;
				case VIDEO_TYPE_VCRYPT:
					decode_frame_vc(tvi.dst,di.width,addr,TV_XSIZE/2,TV_YSIZE/2);
					break;
				default:
					copy_interlace_small(tvi.dst,di.width,addr,TV_XSIZE/2,TV_YSIZE);
					break;
			}
		}

#ifdef PROFILE
		ti=clock() - ti;
		fprintf(stderr,"time=%0.3fs\n",(double) ti / CLOCKS_PER_SEC);

#endif

		/* status display */

		if (disp_status_toggle) {
			DisplayStatus(&di,&tvi);
		} else if (disp_status_delay) {
			disp_status_delay--;
			DisplayStatus(&di,&tvi);
		}

		buffer_nb=1-buffer_nb;
	}

the_end:
	/* kill the audio decoding process */
	set_audio_type(&di,&tvi,&si,AUDIO_TYPE_NORMAL);
	XF86DGADirectVideo(dpy,XDefaultScreen(dpy),0);
	XUngrabPointer(dpy, CurrentTime);
	XUngrabKeyboard(dpy, CurrentTime);
	XFlush(dpy);
	video_mode_close(&tvi,dpy);
	mixer_close(&si);
	grab_close(&gi);
	return 0;
}
