#define VIDEO_TYPE_INTERLACE 0
#define VIDEO_TYPE_PROGRESSIVE 1
#define VIDEO_TYPE_NAGRA 2
#define VIDEO_TYPE_VCRYPT 3

#define AUDIO_TYPE_NONE   0
#define AUDIO_TYPE_NORMAL 1
#define AUDIO_TYPE_NAGRA  2

#define TV_XSIZE_SMALL 384
#define TV_XSIZE  768
#define TV_YSIZE       576

typedef struct CHANNEL {
  char  *name;
  char  *key;
  
  int   capture;
  int   source;
  int   norm;
  int   channel;
  int   fine;
  
  int   color;
  int   bright;
  int   hue;
  int   contrast;

  int video_type;
  int audio_type;
  int gray_scale;

  int volume;
} Channel;

typedef struct STRTAB {
    int  nr;
    char *str;
} StrTab;

typedef struct {
  unsigned char *buffer0;
  unsigned char *buffer1;
  int fd;
  StrTab *inputs;
  StrTab *norms;
  int gb_grab,gb_sync;
} GrabInfo;

typedef struct {
  int mixer_fd;
  int volume;
  int volume_device;
} SoundInfo;

typedef struct {
  int chan_tab;
  int channel_count;
  Channel **channels;

  int cur_color;
  int cur_bright;
  int cur_hue;
  int cur_contrast;
  int cur_input;
  int cur_norm;

  int cur_channel_num;
  int cur_channel;
  int cur_fine;

  int full_screen;
  int cur_video_type;
  int cur_audio_type;
  int cur_gray_scale;
  unsigned char *nagra_table;

  int cur_volume;

  /* grabing structure */
  struct video_mmap *gb;
  /* destination address on screen */
  unsigned char *dst;
  /* position of the TV image */
  int posx,posy;
  int xsize,ysize;
  /* video mode width & height (0 if not used) */
  int vidmode_width;
  int vidmode_height;
} TVInfo;

extern int debug;

/* v4l.c */

int grab_open(GrabInfo *gi,char *filename, int sw, int sh, int bpp);
void grab_close(GrabInfo *gi);

int grab_tune(GrabInfo *gi,unsigned long freq);
int grab_input(GrabInfo *gi,int input, int norm);
int grab_picture(GrabInfo *gi,int color, int bright, int hue, int contrast);
int grab_audio(GrabInfo *gi,int mute, int volume, int *mode);
int grab_frame(GrabInfo *gi,struct video_mmap *gb);


/* config.c */

extern StrTab audiotab[];
extern StrTab videotab[];

void int_to_str(char *buf,int c,StrTab *tab);
int read_config(GrabInfo *gi,TVInfo *tvi,SoundInfo *si,char *filename);
int cf2freq(TVInfo *tvi,int channel, int fine);


/* disp.c */

typedef struct {
  Display *dpy;
  unsigned char *base;
  int width,height,vp_width;
} DispInfo;

void DisplayBitmap(DispInfo *d,
                   int x,int y,int xsize,int ysize,
                   unsigned char *bitmap);

void DisplayText(DispInfo *d,int x,int y,unsigned char *buf,int buf_len);

void ClearScreen(DispInfo *d);

#define FONT_XSIZE 8
#define FONT_YSIZE 14

/* sound.c */

int mixer_open(SoundInfo *si,char *filename);
int mixer_set_device(SoundInfo *si,char *device);
void mixer_set_volume(SoundInfo *si,int volume);
void mixer_close(SoundInfo *si);

void mixer_set_normal(SoundInfo *si);
void mixer_set_decode(SoundInfo *si);

void do_audio_decode(void);
