segment .text

global putline0
global putline1
global putline2
global putline3
global putlinesw16
global putlinesw15
global putlinenocolor0
global putlinenocolor1
global putlinenocolor2
global putlinenocolor3

%macro start_color 0
	push ebp 
	mov ebp,esp
	sub sp,0x40
	push ebx
	push esi
	push ecx
	mov ebx,[ebp+8]        ; src
	mov esi,[ebp+12]       ; dest
	mov ecx,[ebp+16]       ; size 
	shr ecx,2
%endmacro

%macro start_nocolor 0
	push ebp 
	mov ebp,esp 
	sub sp,0x40
	push ebx
	push esi
	push ecx
	push eax
	mov ebx,[ebp+8]        ; src_y
	mov eax,[ebp+12]       ; src_uv
	mov esi,[ebp+16]       ; dest
	mov ecx,[ebp+20]       ; size 
	shr ecx,2
%endmacro


%macro prepconst 0
	movq mm4,[andy]
	movq mm5,[andu]
	movq mm6,[andv]
	movq mm7,[minus128];
%endmacro

%macro loadyuv_color 0
	movq mm1,[ebx]		; u1 y1 v1 y2 u2 y3 v2 y4
	movq mm2,mm1		; u1 y1 v1 y2 u2 y3 v2 y3
	pand mm1,mm4   		; 00 y1 00 y2 00 y3 00 y3
	psrlw mm2,8             ; 00 u1 00 v1 00 u2 00 v2
%endmacro

%macro loadyuv_nocolor 0
	movq mm1,[ebx]		; u1 y1 v1 y2 u2 y3 v2 y4
	movq mm2,[eax]		; u1 y1 v1 y2 u2 y3 v2 y3
	pand mm1,mm4   		; 00 y1 00 y2 00 y3 00 y3
	psrlw mm2,8             ; 00 u1 00 v1 00 u2 00 v2
%endmacro

%macro expandyuv 0
	movq mm3,mm2		; 00 u1 00 v1 00 u2 00 v2
	movq mm0,mm2            ;
	pslld mm3,16		; 00 00 00 u1 00 00 00 u2
	pand mm0,mm6   		; 00 u1 00 00 00 u2 00 00
	por mm3,mm0		; 00 u1 00 u1 00 u2 00 u2 
	
	movq mm0,mm2		; 00 u1 00 v1 00 u2 00 v2
	psrld mm0,16		; 00 v1 00 00 00 v2 00 00
	pand mm2,mm5   		; 00 00 00 v1 00 00 00 v1
	por mm2,mm0             ; 00 v1 00 v1 00 v2 00 v2
%endmacro

%macro expandyvu 0
	movq mm3,mm2		; 00 u1 00 v1 00 u2 00 v2
	movq mm0,mm2            ;
	psrld mm3,16		; 00 00 00 u1 00 00 00 u2
	pand mm0,mm5   		; 00 u1 00 00 00 u2 00 00
	por mm3,mm0		; 00 u1 00 u1 00 u2 00 u2 
	
	movq mm0,mm2		; 00 u1 00 v1 00 u2 00 v2
	pslld mm0,16		; 00 v1 00 00 00 v2 00 00
	pand mm2,mm6   		; 00 00 00 v1 00 00 00 v1
	por mm2,mm0             ; 00 v1 00 v1 00 v2 00 v2
%endmacro

%macro invuv 0
        pandn mm2,mm4		; v-=v
        pandn mm3,mm4		; u-=u
%endmacro

%macro yuv2rgb 0
	; y : mm1
	; v : mm2
	; u : mm3
	paddusw mm2, mm1	; r=v+y;
	paddusw mm3, mm1	; b=u+y;

	psubusw mm2, mm7        ; r=r-128;
        psubusw mm3, mm7        ; b=b-128;

	packuswb mm2,mm2
	punpcklbw mm2,mm2
	pand mm2,mm4  

	packuswb mm3,mm3
	punpcklbw mm3,mm3
	pand mm3,mm4  

	movq mm0, mm1		; 
	psrlw mm0, 2            ; y/4
	psllw mm1, 1		; 2y
	psubusw mm1, mm0	; 1.75y

	psrlw mm2,1		; r/2
	psubusw mm1, mm2	; 1.75y-0.5r

	psrlw mm3,2		; b/4
	psubusw mm1, mm3	; g=1.75y-0.5r-0.25b

	packuswb mm1,mm1
	punpcklbw mm1,mm1
	pand mm1,mm4  
%endmacro
	
%macro rgb16 0
	psrlw mm3,1		; b/4 >> 1
	psrlw mm2,2		; r/2 >> 2
	psllw mm2,11
	psrlw mm1,2
	psllw mm1,5
	por mm1,mm2
	por mm1,mm3
	movq [esi],mm1		; ->
%endmacro

%macro rgb15 0
	psrlw mm3,1		; b/4 >> 1
	psrlw mm2,2		; r/2 >> 2
	psllw mm2,10
	psrlw mm1,3
	psllw mm1,5
	por mm1,mm2
	por mm1,mm3
	movq [esi],mm1		; ->
%endmacro

%macro endcolor 0
	pop ecx
	pop esi
	pop ebx
	emms
	leave
	ret
%endmacro

%macro endnocolor 0
	pop eax
	pop ecx
	pop esi
	pop ebx
	emms
	leave
	ret
%endmacro

putline0:
	start_color
	prepconst
	align 16
.cont0
	loadyuv_color
	expandyuv
	yuv2rgb
	rgb16

	add esi,8
	add ebx,8
	dec ecx
	jnz near .cont0
        endcolor

putline1:
	start_color
	prepconst

	align 16
.cont1
	loadyuv_color
	expandyvu
	invuv
	yuv2rgb
	rgb16

	add esi,8
	add ebx,8
	dec ecx
	jnz near .cont1

	endcolor


putline2:
	start_color
	prepconst

	align 16
.cont2
	loadyuv_color
	expandyuv
	invuv
	yuv2rgb
	rgb16

	add esi,8
	add ebx,8
	dec ecx
	jnz near .cont2

	endcolor


putline3:
	start_color
	prepconst

	align 16
.cont3
	loadyuv_color
	expandyvu
	yuv2rgb
	rgb16

	add esi,8
	add ebx,8
	dec ecx
	jnz near .cont3

	endcolor


putlinenocolor0:
	start_nocolor
	prepconst
	align 16
.cont10
	loadyuv_nocolor
	expandyuv
	yuv2rgb
	rgb16

	add esi,8
	add ebx,8
	add eax,8
	dec ecx
	jnz near .cont10

	endnocolor

putlinenocolor1:
	start_nocolor
	prepconst
	align 16
.cont11
	loadyuv_nocolor
	expandyvu
	invuv
	yuv2rgb
	rgb16

	add esi,8
	add ebx,8
	add eax,8
	dec ecx
	jnz near .cont11

	endnocolor

putlinenocolor2:
	start_nocolor
	prepconst
	align 16
.cont12
	loadyuv_nocolor
	expandyuv
	invuv
	yuv2rgb
	rgb16

	add esi,8
	add ebx,8
	add eax,8
	dec ecx
	jnz near .cont12

	endnocolor


putlinenocolor3:
	start_nocolor
	prepconst
	align 16
.cont13
	loadyuv_nocolor
	expandyvu
	yuv2rgb
	rgb16

	add esi,8
	add ebx,8
	add eax,8
	dec ecx
	jnz near .cont13

	endnocolor

putlinesw16:
	start_color

	movq mm6,[ands]
	align 16
.contsw
	movq mm1,[ebx]
        pand  mm1,mm6	
	psrlw mm1,3
	movq mm2,mm1
	psllw mm2,6
	movq mm3,mm1
	psllw mm3,11
	por mm1,mm2
	por mm1,mm3
	movq [esi],mm1
	add esi,8
	add ebx,8
	dec ecx
	jnz .contsw
	endcolor

putlinesw15:
	start_color

	movq mm6,[ands]
	align 16
.contsw
	movq mm1,[ebx]
        pand  mm1,mm6	
	psrlw mm1,3
	movq mm2,mm1
	psllw mm2,5
	movq mm3,mm1
	psllw mm3,10
	por mm1,mm2
	por mm1,mm3
	movq [esi],mm1
	add esi,8
	add ebx,8
	dec ecx
	jnz .contsw
	endcolor

ands         	dd 0x00f800f8, 0x00f800f8
andy		dd 0x00ff00ff, 0x00ff00ff
andu		dd 0x00ff0000, 0x00ff0000
andv		dd 0x000000ff, 0x000000ff
minus128	dd 0x00800080, 0x00800080
