#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>

#include <sys/soundcard.h>
#include <X11/Xlib.h>

#include "nagratv.h"

static char *names[] = SOUND_DEVICE_NAMES;

int mixer_open(SoundInfo *si,char *filename)
{
  si->mixer_fd=open(filename,O_RDWR);
  if (si->mixer_fd < 0) {
    perror("audio: open mixer");
    si->mixer_fd=-1;
  }
  si->volume_device=SOUND_MIXER_LINE;
  return 0;
}

void mixer_close(SoundInfo *si)
{
  if (si->mixer_fd != -1) {
    close(si->mixer_fd);
  }
}

int mixer_set_device(SoundInfo *si,char *device)
{
  int i,devmask;

  if (-1 == ioctl(si->mixer_fd,MIXER_READ(SOUND_MIXER_DEVMASK),&devmask)) {
    perror("mixer read devmask");
    return -1;
  }

  for (i = 0; i < SOUND_MIXER_NRDEVICES; i++) {
    if ((1<<i) & devmask && strcasecmp(names[i],device) == 0) {
      si->volume_device=i;
      return 0;
    }
  }
  return -1;
}

void mixer_set_volume(SoundInfo *si,int volume)
{
  int tmp;

  if (si->mixer_fd != -1) {
    si->volume=volume;
    tmp=volume;
    tmp=(tmp<<8) | tmp;
    if (ioctl(si->mixer_fd, MIXER_WRITE(si->volume_device), &tmp) == -1) {
      perror("mixer: write");
    }
  }
}

void mixer_set_normal(SoundInfo *si)
{
  /* nothing to do */
}


void mixer_set_decode(SoundInfo *si)
{
  int tmp,chan;

  /* The line should be muted. I don't know how to do it with OSS !!
   Here the volume is set to a low value, but this is not optimal
   because you can hear the crypted sound slightly. */
  tmp=60;
  tmp=(tmp<<8) | tmp;
  if (ioctl(si->mixer_fd, MIXER_WRITE(SOUND_MIXER_LINE), &tmp) == -1) {
    perror("mixer: line");
  }

  /* select line */
	chan = 1 << SOUND_MIXER_LINE;
  ioctl(si->mixer_fd, SOUND_MIXER_WRITE_RECSRC, &chan);

  /* input gain */
  tmp=100;
  tmp=(tmp<<8) | tmp;
  if (ioctl(si->mixer_fd, MIXER_WRITE(SOUND_MIXER_IGAIN), &tmp) == -1) {
    perror("mixer: igain");
  }

  /* pcm volume */
  tmp=100;
  tmp=(tmp<<8) | tmp;
  if (ioctl(si->mixer_fd, MIXER_WRITE(SOUND_MIXER_PCM), &tmp) == -1) {
    perror("mixer: pcm");
  }
}

/* delay in BYTES */
#define DELAY 8192

void do_audio_decode(void)
{
  int i,fd,speed,format,caps,tmp;
  unsigned char *buffer;
  int frag=0xffff000c; /* fragment size of 4k */
  int err;

  /* read device */

  fd=open("/dev/dsp",O_RDWR);
  if (fd < 0) {
    perror("audio: open dsp");
    exit(1);
  }

  /* prepare full duplex mode */

  if (ioctl(fd, SNDCTL_DSP_SETDUPLEX, 0)==-1) {
    perror("/dev/dsp");
    fprintf(stderr, "Error with full duplex setup\n");
    exit(-1);
  }

  if (ioctl(fd, SNDCTL_DSP_GETCAPS, &caps)==-1) {
    perror("/dev/dsp");
    fprintf(stderr, "Sorry but your sound driver is too old\n");
    exit(-1);
  }

  if (!(caps & DSP_CAP_DUPLEX)) {
      fprintf(stderr, "Sorry but your soundcard can't do full duplex\n");
      exit(-1);
  }

  ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &frag);

  /* we use the 8 bit format in order to be compatible 
     with the OSS/Linux SB16 driver */
  format=AFMT_U8;
  if (ioctl(fd,SNDCTL_DSP_SETFMT, &format) == -1) {
    perror("audio: set format");
    exit(1);
  }
  
  tmp=0;
  if (ioctl(fd,SNDCTL_DSP_STEREO, &tmp) == -1) {
    perror("audio: set format");
    exit(1);
  }

  speed=25600;
  if (ioctl(fd,SNDCTL_DSP_SPEED, &speed) == -1) {
    perror("audio: set format");
    exit(1);
  }

  buffer=malloc(DELAY);
  memset(buffer,0x80,DELAY);

  err=write(fd,buffer,DELAY);
  if (err < 0) {
    perror("write");
  }

  while (1) {
    int len;

    len=read(fd,buffer,DELAY);
    if (len < 0) {
      perror("read");
    }
    
    for(i=0;i<DELAY;i+=2) {
      buffer[i]=256-buffer[i];
    }

    err=write(fd,buffer,DELAY);
    if (err < 0) {
      perror("write");
    }
  }
}
