#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <signal.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <time.h>

#include <bttv.h>

#include <X11/Xlib.h>
#include "nagratv.h"

static char *device_cap[] = {
    "capture", "tuner", "teletext", "overlay", "chromakey", "clipping",
    "frameram", "scales", "monochrome", NULL
};

static char *device_pal[] = {
    "-", "grey", "hi240", "rgb565", "rgb24", "rgb32", "rgb555", NULL
};

static struct STRTAB norms[] = {
    {  0, "PAL" },
    {  1, "NTSC" },
    {  2, "SECAM" },
    {  3, "AUTO" },
    { -1, NULL }
};

static struct video_capability  capability;
static struct video_channel     *channels;
static struct video_audio       *audios;
static struct video_tuner       tuner;
static struct video_picture     pict;
static struct video_buffer      ov_fbuf;
static int fd=-1;

unsigned char *cap_buffer;
int swidth,sheight,sbpp;

#define VIDEO_RGB08          1
#define VIDEO_RGB15          2
#define VIDEO_RGB16          4
#define VIDEO_RGB24          8
#define VIDEO_RGB32         16

int grab_open(GrabInfo *gi,char *filename, int sw, int sh, int bpp)
{
  int i,j;

  if (-1 != fd)
    goto err;

  if (-1 == (fd = open(filename ? filename : "/dev/bttv",O_RDWR)))
    goto err;

  if (-1 == ioctl(fd,VIDIOCGCAP,&capability))
    goto err;

  if (debug)
    fprintf(stderr, "v4l: open\n");

  swidth  = sw;
  sheight = sh;
  sbpp = bpp;

  if (debug) {
    fprintf(stderr,"%s:",capability.name);
    for (i = 0; device_cap[i] != NULL; i++)
	    if (capability.type & (1 << i))
        fprintf(stderr," %s",device_cap[i]);
    fprintf(stderr,"\n");
  }

  /* input sources */
  if (debug)
    fprintf(stderr,"  channels: %d\n",capability.channels);
  channels = malloc(sizeof(struct video_channel)*capability.channels);
  memset(channels,0,sizeof(struct video_channel)*capability.channels);
  gi->inputs = malloc(sizeof(struct STRTAB)*(capability.channels+1));
  memset(channels,0,sizeof(struct STRTAB)*(capability.channels+1));
  for (i = 0; i < capability.channels; i++) {
    channels[i].channel = i;
    if (-1 == ioctl(fd,VIDIOCGCHAN,&channels[i]))
	    perror("ioctl VIDIOCGCHAN"), exit(0);
    gi->inputs[i].nr  = i;
    gi->inputs[i].str = channels[i].name;
    if (debug)
	    fprintf(stderr,"    %s: %d %s%s %s%s\n",
		    channels[i].name,
		    channels[i].tuners,
		    (channels[i].flags & VIDEO_VC_TUNER)   ? "tuner "  : "",
		    (channels[i].flags & VIDEO_VC_AUDIO)   ? "audio "  : "",
		    (channels[i].type & VIDEO_TYPE_TV)     ? "tv "     : "",
		    (channels[i].type & VIDEO_TYPE_CAMERA) ? "camera " : "");

    /* we suppose there is only ONE tuner for all the channels */
    if (channels[i].flags & VIDEO_VC_TUNER) {
    
      tuner.tuner=0;
      if (-1 == ioctl(fd,VIDIOCGTUNER,&tuner))
        perror("ioctl VIDIOCGTUNER");
      /* bug fix of the driver for SECAM ! */
      tuner.flags |= VIDEO_TUNER_SECAM;
      if (debug)
        fprintf(stderr,"  tuner   : %s %lu-%lu",
                tuner.name,tuner.rangelow,tuner.rangehigh);
      for (j = 0; norms[j].str != NULL; j++) {
        if (tuner.flags & (1<<j)) {
          if (debug)
            fprintf(stderr," %s",norms[j].str);
        }
        if (tuner.mode == j && debug)
          fprintf(stderr,"*");
      }
      if (debug)
        fprintf(stderr,"\n");
      gi->norms=norms;
    }
  }
  gi->inputs[i].nr  = -1;
  gi->inputs[i].str = NULL;

    /* audios */
    if (debug)
	fprintf(stderr,"  audios  : %d\n",capability.audios);
    audios = malloc(sizeof(struct video_audio)*capability.audios);
    memset(audios,0,sizeof(struct video_audio)*capability.audios);
    for (i = 0; i < capability.audios; i++) {
	audios[i].audio = i;
	if (-1 == ioctl(fd,VIDIOCGAUDIO,&audios[i]))
	    perror("ioctl VIDIOCGCAUDIO") /* , exit(0) */ ;
	if (debug) {
	    fprintf(stderr,"    %d (%s): ",i,audios[i].name);
	    if (audios[i].flags & VIDEO_AUDIO_MUTABLE)
		fprintf(stderr,"muted=%s ",
			(audios[i].flags&VIDEO_AUDIO_MUTE) ? "yes":"no");
	    if (audios[i].flags & VIDEO_AUDIO_VOLUME)
		fprintf(stderr,"volume=%d ",audios[i].volume);
	    if (audios[i].flags & VIDEO_AUDIO_BASS)
		fprintf(stderr,"bass=%d ",audios[i].bass);
	    if (audios[i].flags & VIDEO_AUDIO_TREBLE)
		fprintf(stderr,"treble=%d ",audios[i].treble);
	    fprintf(stderr,"\n");
	}
    }

    if (debug)
	fprintf(stderr,"  size    : %dx%d => %dx%d\n",
		capability.minwidth,capability.minheight,
		capability.maxwidth,capability.maxheight);


    /* frame buffer */
    if (-1 == ioctl(fd,VIDIOCGFBUF,&ov_fbuf))
	perror("ioctl VIDIOCGFBUF");
    if (debug)
	fprintf(stderr,"  fbuffer : base=0x%p size=%dx%d depth=%d bpl=%d\n",
		ov_fbuf.base, ov_fbuf.width, ov_fbuf.height,
		ov_fbuf.depth, ov_fbuf.bytesperline);

    /* picture parameters */
    if (-1 == ioctl(fd,VIDIOCGPICT,&pict))
	perror("ioctl VIDIOCGPICT");

    if (debug) {
	fprintf(stderr,
		"  picture : brightness=%d hue=%d colour=%d contrast=%d\n",
		pict.brightness, pict.hue, pict.colour, pict.contrast);
	fprintf(stderr,
		"  picture : whiteness=%d depth=%d palette=%s\n",
		pict.whiteness, pict.depth, device_pal[pict.palette]);
    }

    if (0 != strncmp(capability.name,"BT848",5)) goto err;
    
    gi->buffer0 = mmap(0,2*BTTV_MAX_FBUF,
                   PROT_READ|PROT_WRITE,MAP_SHARED,fd,0);
    if ((unsigned char*)-1 == gi->buffer0) {
      perror("mmap");
      goto err;
    }

    gi->buffer1 = gi->buffer0 + BTTV_MAX_FBUF;

    gi->fd=fd;

    gi->gb_grab=0;
    gi->gb_sync=0;

    return 0;

err:
    if (fd != -1) {
	close(fd);
	fd = -1;
    }
    return -1;
}

void grab_close(GrabInfo *gi)
{
  /* stay synchronised */

#if 0
  while (gi->gb_grab > gi->gb_sync) {
    if (-1 == ioctl(fd,VIDIOCSYNC,0)) {
      if (errno != EAGAIN)
          perror("ioctl VIDIOCSYNC");
    } else
	    gi->gb_sync++;
  }
#endif
  
  if (gi->buffer0 != NULL) munmap(gi->buffer0,2*swidth*sheight*sbpp);

  if (gi->fd != -1) {
    if (debug)
      fprintf(stderr, "v4l: close\n");

    close(gi->fd);
  }
}

/* Grab one frame. The current video4linux grab API is horrible ! */

int grab_frame(GrabInfo *gi,struct video_mmap *gb)
{

  if (gi->gb_grab == gi->gb_sync) {
    if (-1 == ioctl(fd,VIDIOCMCAPTURE,gb)) {
	    if (errno == EAGAIN) /* bttv: no station tuned in */
		    return ioctl (fd, BTTV_PICNR,0);
      else
        perror("ioctl VIDIOCMCAPTURE");
    } else {
	    gi->gb_grab++;
    }
  }

  if (-1 == ioctl(fd,VIDIOCMCAPTURE,gb)) {
    if (errno == EAGAIN) /* bttv: no station tuned in */
	    return ioctl (fd, BTTV_PICNR,0);
    else
      perror("ioctl VIDIOCMCAPTURE");
  } else {
    gi->gb_grab++;
  }

  if (-1 == ioctl(fd,VIDIOCSYNC,0)) {
    if (errno != EAGAIN) 
      perror("ioctl VIDIOCSYNC");
  } else
    gi->gb_sync++;
  return ioctl (fd, BTTV_PICNR,0);
}

int grab_tune(GrabInfo *gi,unsigned long freq)
{
    if (debug)
	fprintf(stderr,"v4l: freq: %.3f\n",(float)freq/16);
    if (-1 == ioctl(fd, VIDIOCSFREQ, &freq))
	perror("ioctl VIDIOCSFREQ");
    return 0;
}

int grab_input(GrabInfo *gi,int input, int norm)
{
  if (-1 != input) {
    if (debug)
	    fprintf(stderr,"v4l: input: %d\n",input);
    if (-1 == ioctl(fd, VIDIOCSCHAN, &input))
	    perror("ioctl VIDIOCSCHAN");
  }

  if (-1 != norm) {
    if (debug)
	    fprintf(stderr,"v4l: norm : %d\n",norm);
    tuner.mode = norm;
    if (ioctl(fd, VIDIOCSTUNER, &tuner) < 0)
	    perror("ioctl VIDIOCSTUNER");

    if (ioctl(fd, VIDIOCGTUNER, &tuner) < 0)
	    perror("ioctl VIDIOCSTUNER");
    if (debug)
      printf("tuner.mode=%d\ntuner.flags=%d\n",tuner.mode,tuner.flags);
  }
  return 0;
}

int grab_picture(GrabInfo *gi,int color, int bright, int hue, int contrast)
{
    if (color != -1)
	pict.colour = color;
    if (contrast != -1)
	pict.contrast = contrast;
    if (bright != -1)
	pict.brightness = bright;
    if (hue != -1)
	pict.hue = hue;

    if (-1 == ioctl(fd,VIDIOCSPICT,&pict))
	perror("ioctl VIDIOCSPICT");

    return 0;
}

int grab_audio(GrabInfo *gi,int mute, int volume, int *mode)
{
    if (mute != -1) {
	if (mute)
	    audios[0].flags |= VIDEO_AUDIO_MUTE;
	else
	    audios[0].flags &= ~VIDEO_AUDIO_MUTE;
    }
    if (volume != -1)
	audios[0].volume = volume;

    audios[0].mode = mode ? *mode : 0;
    if (-1 == ioctl(fd,VIDIOCSAUDIO,&audios[0]))
	perror("ioctl VIDIOCSAUDIO");

    if (mode) {
	if (-1 == ioctl(fd,VIDIOCGAUDIO,&audios[0]))
	    perror("ioctl VIDIOCGAUDIO");
	*mode = audios[0].mode;
    }
    return 0;
}
